#include <iostream>
#include <cstring>
#include <cctype>
#include <unistd.h>
#include <SDL.h>
#include "xray/xray.hpp"
#include "common/timer.h"
#include "common/err_msg.h"

int Init(int x, int y);
void UpdateGfx();
void CleanUp();
void OutputRenderStats(xray::RenderStats *stats);
void HandleEvent(SDL_Event &event);
void ProgressBar(float progress);

static SDL_Surface *fb;
static xray::Scene *scene;

static int x = 800;
static int y = 600;

static SDL_Rect border_out_rect, border_in_rect, full_progr_rect;
static bool done = false;

#define MAX_FNAME_SIZE	512
char fname[MAX_FNAME_SIZE] = "scene.xml";


int main(int argc, char **argv) {

	for(int i=1; i<argc; i++) {
		if(argv[i][0] == '-') {
			if(!strcmp(argv[i], "-size")) {
				char *size = strdup(argv[++i]);
				char *sep = strchr(size, 'x');
				if(!sep || !*(sep+1) || !isdigit(size[0]) || !isdigit(*(sep+1))) {
					std::cerr << "malformed resolution, use form <x-res>x<y-res> (e.g: 800x600)\n";
					return -1;
				}
				x = atoi(size);
				y = atoi(sep+1);
			}
		} else {
			strncpy(fname, argv[i], MAX_FNAME_SIZE-1);
			fname[MAX_FNAME_SIZE-1] = 0;
		}
	}
	
	if(Init(x, y) == -1) return -1;

	while(!done) {
		SDL_Event event;
		if(!SDL_WaitEvent(&event)) {
			std::cerr << "oug\n";
			break;
		}

		HandleEvent(event);
	}

	CleanUp();

	return done ? 0 : -1;
}

int Init(int x, int y) {
	if(!xray::Init(x, y)) {
		return -1;
	}

	SDL_Init(SDL_INIT_VIDEO);
	
	if(!(fb = SDL_SetVideoMode(x, y, 32, SDL_HWSURFACE | SDL_DOUBLEBUF))) {
		return -1;
	}

	SDL_WM_SetCaption("X-Ray :: a UNIX raytracer", 0);

	// initialize progress bar rects...
	border_out_rect.x = x/2 - x/3;
	border_out_rect.y = y/2 - y/24;
	border_out_rect.w = 2 * x / 3;
	border_out_rect.h = 2 * y / 24;

	border_in_rect = border_out_rect;
	border_in_rect.x += 4;
	border_in_rect.y += 4;
	border_in_rect.w -= 8;
	border_in_rect.h -= 8;

	full_progr_rect = border_in_rect;
	full_progr_rect.x += 4;
	full_progr_rect.y += 4;
	full_progr_rect.w -= 8;
	full_progr_rect.h -= 8;

	using namespace xray;

	// initialize scene
	scene = LoadScene(fname);
	if(!scene) {
		std::cerr << "scene load oug: " << fname << std::endl;
		return -1;
	}
	
	SetProgressCallback(ProgressBar);

	std::cout << "Rendering... \n";
	UpdateGfx();


	return 0;
}

void UpdateGfx() {
	//SDL_FillRect(fb, 0, 0x00ff0000);
	
	ntimer timer;
	timer_reset(&timer);
	
	int i = 0;
	while(!done && scene->RenderScanline(i++)) {
		SDL_Event event;
		while(SDL_PollEvent(&event)) {
			HandleEvent(event);
		}
	}
	done = false;

	unsigned long rtime = timer_getmsec(&timer);
	info("rendering time: %.2f sec (%lu msec)", (float)rtime / 1000.0f, rtime);

	xray::RenderStats *stats = xray::GetRenderStats();
	OutputRenderStats(stats);
	
	if(SDL_MUSTLOCK(fb)) SDL_LockSurface(fb);

	const PixelBuffer *pbuf = xray::GetPixelBuffer();
	memcpy(fb->pixels, pbuf->buffer, pbuf->width * pbuf->height * sizeof(Pixel));

	if(SDL_MUSTLOCK(fb)) SDL_UnlockSurface(fb);

	SDL_Flip(fb);
	usleep(10000);
}

void CleanUp() {
	xray::Destroy();
	delete scene;
}

void OutputRenderStats(xray::RenderStats *stats) {
	info("---- Rendering Statistics -----");
	info("   Primary rays: %lu", stats->primary_rays);
	info(" Secondary rays: %lu", stats->secondary_rays);
	info("    Shadow rays: %lu", stats->shadow_rays);
	info("Glossiness rays: %lu", stats->gloss_rays);
}

void HandleEvent(SDL_Event &event) {
	switch(event.type) {
	case SDL_QUIT:
		done = true;
		break;

	case SDL_KEYDOWN:
		if(event.key.keysym.sym == SDLK_ESCAPE) {
			done = true;
		}
		break;
	}
}

void ProgressBar(float progress) {
	//SDL_FillRect(fb, &border_out_rect, 0);

	SDL_FillRect(fb, &border_out_rect, 0x00ff0000);
	SDL_FillRect(fb, &border_in_rect, 0);

	SDL_Rect progr_rect = full_progr_rect;
	progr_rect.w = (int)((float)progr_rect.w * progress);
	SDL_FillRect(fb, &progr_rect, 0x00ff0000);

	SDL_Flip(fb);
}
